/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file config_parse.c
 * Implementation of the Ceph/S3 DSI
 */

#include "globus_gridftp_server_s3_base.h"

static
globus_result_t
globus_l_s3_config_parse_key(
    const char                         *config_path,
    int                                 lineno,
    const char                         *key,
    const char                         *val,
    globus_i_s3_config_key_t           *config_keys,
    bool                               *parsed_option);

static
globus_result_t
globus_l_s3_config_check_required(
    const char                         *path,
    const globus_i_s3_config_key_t     *config_keys);

/**
 * @brief Parse configuration file
 * @details
 *      Parse the configuration file located at config_path. If
 *      extra_config_keys is not NULL, those keys, as well as the default
 *      keys will be parsed.
 *
 * @param config_path
 * @param extra_config_keys
 */
globus_result_t
globus_i_s3_config_parse(
    const char                         *config_path,
    globus_i_s3_config_key_t           *extra_config_keys)
{
    char                                buf[256];
    char                                key[256];
    char                                val[256];
    int                                 lineno = 0;
    globus_result_t                     result = GLOBUS_SUCCESS;
    FILE                               *fp = NULL;
    globus_i_s3_config_key_t            default_config_keys[] =
    {
        {
            .value_type = GLOBUS_S3_CONFIG_STRING,
            .key_name = "host_name",
            .value_string = &globus_i_s3_default_hostname
        },
        {
            .value_type = GLOBUS_S3_CONFIG_STRING,
            .key_name = "mime_types_path",
            .value_string = &globus_i_s3_mime_type_file
        },
        {
            .value_type = GLOBUS_S3_CONFIG_BOOL,
            .key_name = "use_https",
            .value_bool = &globus_i_s3_https
        },
        {
            .value_type = GLOBUS_S3_CONFIG_BOOL,
            .key_name = "use_virtual_host",
            .value_bool = &globus_i_s3_use_virtual_host
        },
        {
            .value_type = GLOBUS_S3_CONFIG_BOOL,
            .key_name = "verify_peer",
            .value_bool = &globus_i_s3_verify_peer
        },
        {
            .value_type = GLOBUS_S3_CONFIG_INT,
            .key_name = "max_retries",
            .value_int = &globus_i_s3_max_retries
        },
        {
            .value_type = GLOBUS_S3_CONFIG_INT,
            .key_name = "retry_delay",
            .value_int = &globus_i_s3_retry_delay_ms
        },
        {
            .key_name = NULL
        }

    };

    GlobusS3Enter();

    GlobusS3Trace("parsing config file %s\n", config_path);

    fp = fopen(config_path, "r");
    if (fp == NULL)
    {
        goto done;
    }

    while (fgets(buf, sizeof(buf), fp) != NULL)
    {
        char                           *end;
        char                           *start;
        char                           *comment;
        bool                            parsed_option = false;
        int                             rc;

        lineno++;

        GlobusS3Trace("parsing line [%d]: %s\n", lineno, buf);

        if ((comment = strchr(buf, '#')) != NULL)
        {
            *comment = 0;
        }
        start = buf;

        while (isspace(*start))
        {
            start++;
        }
        if (*start == 0)
        {
            continue;
        }
        end = buf + strlen(buf)-1;
        while (isspace(*end))
        {
            *end = 0;
            end--;
        }
        rc = sscanf(start, "%s = %s", key, val);
        if (rc != 2)
        {
            const char *fmt = "Error parsing %s:%d: invalid line\n";
            size_t msg_len = snprintf(NULL, 0, fmt, config_path, lineno);
            char msg[msg_len+1];

            snprintf(msg, sizeof(msg), fmt, config_path, lineno);
            result = GlobusGFSErrorGeneric(msg);

            goto error_parse;
        }
        result = globus_l_s3_config_parse_key(
                config_path, lineno,
                key, val, default_config_keys, &parsed_option);

        if (result != GLOBUS_SUCCESS)
        {
            goto error_parse;
        }
        if (!parsed_option)
        {
            if (extra_config_keys != NULL)
            {
                result = globus_l_s3_config_parse_key(
                        config_path, lineno,
                        key, val, extra_config_keys, &parsed_option);

            }
            if (result != GLOBUS_SUCCESS)
            {
                goto error_parse;
            }
            if (!parsed_option)
            {
                const char *fmt = "Error parsing %s:%d: unknown option \"%s\"\n";
                size_t msg_len = snprintf(NULL, 0, fmt, config_path, lineno, key);
                char msg[msg_len+1];

                snprintf(msg, sizeof(msg), fmt, config_path, lineno, key);

                result = GlobusGFSErrorGeneric(msg);

                goto error_parse;
            }
        }
    }
    result = globus_l_s3_config_check_required(
            config_path,
            default_config_keys);
    if (result == GLOBUS_SUCCESS && extra_config_keys != NULL)
    {
        result = globus_l_s3_config_check_required(
                config_path,
                extra_config_keys);
    }
    if (result != GLOBUS_SUCCESS)
    {
error_parse:
        for (size_t i = 0; default_config_keys[i].key_name != NULL; i++)
        {
            switch (default_config_keys[i].value_type)
            {
                case GLOBUS_S3_CONFIG_STRING:
                    free(*(default_config_keys[i].value_string));
                    *(default_config_keys[i].value_string) = NULL;
                    break;
                default:
                    break;
            }
        }
        if (extra_config_keys != NULL)
        {
            for (size_t i = 0; extra_config_keys[i].key_name != NULL; i++)
            {
                switch (extra_config_keys[i].value_type)
                {
                    case GLOBUS_S3_CONFIG_STRING:
                        free(*(extra_config_keys[i].value_string));
                        *(extra_config_keys[i].value_string) = NULL;
                        break;
                    default:
                        break;
                }
            }
        }
    }
done:
    if (fp != NULL)
    {
        fclose(fp);
    }
    GlobusS3ExitResult(result);
    return result;
}
/* globus_i_s3_config_parse() */

static
globus_result_t
globus_l_s3_config_parse_key(
    const char                         *config_path,
    const int                           lineno,
    const char                         *key,
    const char                         *val,
    globus_i_s3_config_key_t           *config_keys,
    bool                               *parsed_option)
{
    globus_result_t                     result = GLOBUS_SUCCESS;
    char                               *valend;

    GlobusS3Enter();

    assert(key != NULL);
    assert(val != NULL);
    assert(config_keys != NULL);
    assert(parsed_option != NULL);

    for (size_t i = 0; config_keys[i].key_name != NULL; i++)
    {
        const char *key_name = config_keys[i].key_name;

        if (strcmp(key, key_name) == 0)
        {
            switch (config_keys[i].value_type)
            {
                case GLOBUS_S3_CONFIG_STRING:
                    if (*(config_keys[i].value_string))
                    {
                        if (strcmp(*(config_keys[i].value_string), val) != 0)
                        {
                            free(*(config_keys[i].value_string));

                            *(config_keys[i].value_string) = strdup(val);
                        }
                    }
                    else
                    {
                        *(config_keys[i].value_string) = strdup(val);
                    }

                    if (*(config_keys[i].value_string) == NULL)
                    {
                        result = GlobusGFSErrorMemory("value_string");
                        goto config_strdup_fail;
                    }
                    *parsed_option = true;
                    break;
                case GLOBUS_S3_CONFIG_BOOL:
                    if (strcasecmp(val, "true") == 0
                        || strcasecmp(val, "yes") == 0
                        || strcasecmp(val, "on") == 0
                        || strcasecmp(val, "1") == 0)
                    {
                        *(config_keys[i].value_bool) = true;
                    }
                    else if (strcasecmp(val, "false") == 0
                        || strcasecmp(val, "no") == 0
                        || strcasecmp(val, "off") == 0
                        || strcasecmp(val, "0") == 0)
                    {
                        *(config_keys[i].value_bool) = false;
                    }
                    else
                    {
                        const char *fmt = "Error parsing %s:%d: unexpected value for boolean attribute %s: \"%s\"\n";
                        size_t msg_len = snprintf(NULL, 0, fmt, config_path, lineno, config_keys[i].key_name, val);
                        char msg[msg_len+1];

                        snprintf(msg, sizeof(msg), fmt, config_path, lineno, config_keys[i].key_name, val);


                        result = GlobusGFSErrorGeneric(msg);

                        goto error_parse;
                    }
                    *parsed_option = true;
                    break;
                case GLOBUS_S3_CONFIG_INT:
                    errno = 0;
                    valend = NULL;
                    (*config_keys[i].value_int) = (int) strtol(val, &valend, 10);
                    if (errno != 0 || valend == NULL || *valend != 0 || (*config_keys[i].value_int) < 0)
                    {
                        const char *fmt = "Error parsing %s:%d: unexpected value for integer attribute %s: \"%s\"\n";
                        size_t msg_len = snprintf(NULL, 0, fmt, config_path, lineno, config_keys[i].key_name, val);
                        char msg[msg_len+1];

                        snprintf(msg, sizeof(msg), fmt, config_path, lineno, config_keys[i].key_name, val);


                        result = GlobusGFSErrorGeneric(msg);

                        goto error_parse;
                    }
                
                    *parsed_option = true;
                    break;
                default:
                    assert(config_keys[i].value_type == GLOBUS_S3_CONFIG_STRING
                        || config_keys[i].value_type == GLOBUS_S3_CONFIG_BOOL);
            }

            break;
        }
    }

error_parse:
config_strdup_fail:

    GlobusS3ExitResult(result);
    return result;
}
/* globus_l_s3_config_parse_key() */

static
globus_result_t
globus_l_s3_config_check_required(
    const char                         *path,
    const globus_i_s3_config_key_t     *config_keys)
{
    globus_result_t                     result = GLOBUS_SUCCESS;

    for (size_t i = 0; config_keys[i].key_name != NULL; i++)
    {
        if (config_keys[i].required
            && (*config_keys[i].value_string) == NULL)
        {
            const char *err_fmt = "Required config \"%s\" not set in %s\n";
            size_t msg_len = snprintf(NULL, 0, err_fmt, config_keys[i].key_name, path);
            char msg[msg_len+1];

            snprintf(msg, sizeof(msg), err_fmt, config_keys[i].key_name, path);

            GlobusS3Error(msg);

            result = GlobusGFSErrorGeneric(msg);

            goto missing_required;
        }
    }
missing_required:
    return result;
}
/* globus_l_s3_config_check_required() */
// vim: ft=c:
